/*=============================================Libraries===============================================*/

#include <SoftwareSerial.h>
SoftwareSerial ZigBee_serial(0, 1); // 0->RX, 1->TX

/*===========================================Declarations==============================================*/

const int ZigBee_LED = 2;

char HVAC_data[17];
char ZigBee_data[15];
char ZigBee_data_cmd[5];
char ZigBee_data_value[10];
char payload_to_HVAC[23];
char payload_to_ZigBee[60];
char *index_t;

unsigned long start_millis;
unsigned long current_millis;
unsigned long intervall = 2000;

int size_data2;

boolean value_changed = false;
boolean init_state = false;

byte temp_value = 0;

const char mode_list[6][9] = {
  {"off"},      //0x0000
  {"heat"},     //0x0010
  {"dry"},      //0x0020
  {"-"},
  {"cool"},     //0x0040
  {"fan_only"}, //0x0050
};
const char fan_list[4][7] = {
  {"auto"},   //0x00
  {"high"},   //0x01
  {"medium"}, //0x02
  {"low"},    //0x02
};

uint16_t address_list[]=
{
  0x0000, //power (ON, OFF)
  0x0001, //mode (hot, dry, cool, fan)
  0x0002, //fan speed (auto, high, medium, low)
  0x0003, //temperature (16~32°C)
  0x0100  //inside temperature
};

struct HVAC_status {
  byte power;
  byte mode;
  byte fan_speed;
  byte temperature;
  byte inside_temp;
} Hitachi_status;

/*==============================================Interrupt================================================*/

void LED_ISR(){
  noInterrupts(); 
  digitalWrite(LED_BUILTIN, !digitalRead(ZigBee_LED)); 
  interrupts();
}

/*===============================================Setup=================================================*/

void setup() {
 pinMode(LED_BUILTIN, OUTPUT);
 pinMode(ZigBee_LED, INPUT);
 digitalWrite(LED_BUILTIN, !digitalRead(ZigBee_LED));
 attachInterrupt(digitalPinToInterrupt(ZigBee_LED), LED_ISR, CHANGE);  
 ZigBee_serial.begin(9600);
 Serial1.begin(9600, SERIAL_8O1); // baud raute: 9600, data size: 8 bits, parity: odd, stop bits: 1
 size_t a = Serial1.readBytesUntil('\r', HVAC_data, 17); // discard the message [NG P=01 C=FFFE\r] received from HVAC after initialization of Serial1
 delay(20);

 Hitachi_status.power = inquire_HVAC_status(address_list[0]);
 Hitachi_status.mode = inquire_HVAC_status(address_list[1]);
 Hitachi_status.fan_speed = inquire_HVAC_status(address_list[2]);
 Hitachi_status.temperature = inquire_HVAC_status(address_list[3]);
 Hitachi_status.inside_temp = inquire_HVAC_status(address_list[4]);
 
 send_HVAC_status();
 
 start_millis = millis();
}

/*=============================================Main loop===============================================*/

void loop(){
  if(ZigBee_serial.available() > 0) {
    size_data2 = ZigBee_serial.readBytesUntil('\r', ZigBee_data, 15);
    ZigBee_data[size_data2] = '\0';
    index_t = strchr(ZigBee_data, ':');
    *index_t = '\0';
    strcpy(ZigBee_data_cmd, ZigBee_data);
    strcpy(ZigBee_data_value, index_t + 1);
    if(strcmp(ZigBee_data_cmd,"MODE") == 0){
      if(strcmp(ZigBee_data_value, mode_list[0]) == 0){
        Hitachi_status.power = 0x00;
        set_HVAC_status(address_list[0],0x00);
      }
      else{
        for(int i = 1; i < 6; i++){
          if(strcmp(ZigBee_data_value, mode_list[i]) == 0){
            Hitachi_status.mode = i << 4;
            Hitachi_status.power = 0x01;
            set_HVAC_status(address_list[1],Hitachi_status.mode);
            set_HVAC_status(address_list[0],1);
          }
        }
      }
    }
    else if(strcmp(ZigBee_data_cmd,"FAN") == 0){
      for(int j = 0; j < 4; j++){
        if(strcmp(ZigBee_data_value, fan_list[j]) ==0){
          Hitachi_status.fan_speed = j;
          set_HVAC_status(address_list[2],Hitachi_status.fan_speed);
        }
      }
    }
    else if(strcmp(ZigBee_data_cmd,"TMP") == 0){
      index_t = strchr(ZigBee_data_value, '.');
      *index_t = '\0';
      Hitachi_status.temperature = atoi(ZigBee_data_value);
      set_HVAC_status(address_list[3],Hitachi_status.temperature);
    }
    send_HVAC_status();
  }
  
  current_millis = millis();
  if(current_millis - start_millis >= intervall){
    start_millis = millis();
    temp_value = inquire_HVAC_status(address_list[0]);
    if(temp_value != Hitachi_status.power){
      value_changed = true;
      Hitachi_status.power = temp_value;
    }
    temp_value = inquire_HVAC_status(address_list[1]);
    if(temp_value != Hitachi_status.mode){
      value_changed = true;
      Hitachi_status.mode = temp_value;
    }
    temp_value = inquire_HVAC_status(address_list[2]);
    if(temp_value != Hitachi_status.fan_speed){
      value_changed = true;
      Hitachi_status.fan_speed = temp_value;
    }
    temp_value = inquire_HVAC_status(address_list[3]);
    if(temp_value != Hitachi_status.temperature){
      value_changed = true;
      Hitachi_status.temperature = temp_value;
    } 
    temp_value = inquire_HVAC_status(address_list[4]);
    if(temp_value != Hitachi_status.inside_temp){
      value_changed = true;
      Hitachi_status.inside_temp = temp_value;
    }
    if(value_changed == true){
      send_HVAC_status();
      value_changed = false;
    }
    temp_value = 0;
  }

}

/*=============================================Subroutines=============================================*/

/*-----------------------------------------------------------*/
byte inquire_HVAC_status(uint16_t address_to_send){
  int size_data;
  byte cs_buffer;
  byte status_value;
  char buffer1[5];
  char *ptr;

  strcpy(payload_to_HVAC,"MT P=0");
  if(address_to_send == 0x0100){
    strcat(payload_to_HVAC,"100");
  }
  else{
    strcat(payload_to_HVAC,"00");
    itoa(address_to_send, buffer1, 16);
    strcat(payload_to_HVAC, buffer1);
  }
  strcat(payload_to_HVAC," C=FF");
  if(address_to_send == 0x0100){
    cs_buffer = 0xFF - 1;
    }
  else{
    cs_buffer = 0xFF - address_to_send;
    }
  buffer1[0]=0;
  itoa(cs_buffer,buffer1,16);
  strcat(payload_to_HVAC, buffer1);
  strupr(payload_to_HVAC);
  strcat(payload_to_HVAC,"\r");
  Serial1.print(payload_to_HVAC);
  Serial1.flush();
  
  size_data = Serial1.readBytesUntil('\r', HVAC_data, 17);
  ptr = strtok (HVAC_data, "=");
  ptr = strtok (NULL, " ");
  status_value = (byte) strtoul(ptr, NULL, 16);
  delay(20);
  return status_value;
  
}
/*-----------------------------------------------------------*/

/*-----------------------------------------------------------*/
void send_HVAC_status(){
  char convert_buffer2[5];
  strcpy(payload_to_ZigBee, "{\"MODE\":\"");
  if(Hitachi_status.power == 0){
    strcat(payload_to_ZigBee, mode_list[0]);
    }
  else {
    strcat(payload_to_ZigBee, mode_list[Hitachi_status.mode >> 4]);
    }
  strcat(payload_to_ZigBee, "\",\"FAN\":\"");
  strcat(payload_to_ZigBee, fan_list[Hitachi_status.fan_speed]);
  strcat(payload_to_ZigBee, "\",\"TMP\":");
  itoa(Hitachi_status.temperature, convert_buffer2, 10);
  strcat(payload_to_ZigBee,convert_buffer2);
  strcat(payload_to_ZigBee,",\"IN_TMP\":");
  convert_buffer2[0]=0;
  itoa(Hitachi_status.inside_temp, convert_buffer2, 10);
  strcat(payload_to_ZigBee,convert_buffer2);
  strcat(payload_to_ZigBee,"}\r");
  ZigBee_serial.print(payload_to_ZigBee);
  ZigBee_serial.flush();
}
/*-----------------------------------------------------------*/

/*-----------------------------------------------------------*/
void set_HVAC_status(uint16_t address_to_send, byte value_to_send){
  char convert_buffer[5];
  byte value_buffer;
  strcpy(payload_to_HVAC, "ST P=000");
  itoa(address_to_send, convert_buffer, 10);
  strcat(payload_to_HVAC, convert_buffer);
  strcat(payload_to_HVAC, ",0");
  if(address_to_send == 0x0001 or address_to_send == 0x003){
    strcat(payload_to_HVAC, "0");
  }
  convert_buffer[0]=0;
  itoa(value_to_send, convert_buffer, 16);
  strcat(payload_to_HVAC, convert_buffer);
  strcat(payload_to_HVAC, " C=FF");
  value_buffer = 0xFF - address_to_send - value_to_send;
  convert_buffer[0]=0;
  itoa(value_buffer, convert_buffer, 16);
  strcat(payload_to_HVAC, convert_buffer);
  strupr(payload_to_HVAC);
  strcat(payload_to_HVAC, "\r");
  Serial1.print(payload_to_HVAC);
  Serial1.flush();

  size_t z = Serial1.readBytesUntil('\r', HVAC_data, 17);
  delay(20);
}
/*-----------------------------------------------------------*/
